/* --------------------------------------------------------------- */
/* lpcBitIO.c -- bit device functions for LPC                      */
/* --------------------------------------------------------------- */
/*                                                                 */
/* Copyright (c) Mike Cowlishaw, 2010.  All rights reserved.       */
/*                                                                 */
/* Permission to use, copy, modify, and/or distribute this         */
/* software for any purpose with or without fee is hereby granted, */
/* provided that the above copyright notice and this permission    */
/* notice appear in all copies, and that notice and the date of    */
/* any modifications be added to the software.                     */
/*                                                                 */
/* This software is provided "as is".  No warranties, whether      */
/* express, implied, or statutory, including, but not limited to,  */
/* implied warranties of merchantability and fitness for a         */
/* particular purpose apply to this software.  The author shall    */
/* not, in any circumstances, be liable for special, incidental,   */
/* or consequential damages, for any reason whatsoever.            */
/*                                                                 */
/* --------------------------------------------------------------- */
/* These functions take a port (GPIO register) number and a bit    */
/* mask -- all those bits that have a 1 in the mask are affected.  */
/*                                                                 */
/* bitSetIO sets the configuration of a bit; bitIntIO controls     */
/* its interrupt capabilities, and bitExistIO checks whether a     */
/* pin exists in the hardware.                                     */
/* --------------------------------------------------------------- */

#define  TOLLOS_IO                 // allow use of I/O functions
#include "tollosBit.h"
#include "lpc17xx_gpio.h"
#include "lpc17xx_pinsel.h"

// master bit interrupt handler
static void bitHandler(void);

// arrays of handlers; one for each bit that might cause an interrupt
#define BITS0 31
#define BITS2 14
static vector bithandlers0[BITS0]; // for bits 0...30 in PORT0
static vector bithandlers2[BITS2]; // for bits 0...13 in PORT2

// sysEnumVector -- convert LPC17xx.h enum name to vector table index
#define sysEnumVector(e) ((int)(e)+16)

/* --------------------------------------------------------------- */
/* bitExistIO -- test whether a bit or bits exist                  */
/*                                                                 */
/*   port   -- port in which the bit or bits reside                */
/*   mask   -- mask of which bit(s) to test                        */
/*                                                                 */
/*   returns 1 if all the bits in the mask exist in the specified  */
/*     port, or 0 otherwise; 0 is also returned if port is invalid */
/*                                                                 */
/* Note: this reports whether the bit exists in a microcontroller, */
/* not whether it is brought out to a board pin, or whether it     */
/* supports interrupts.                                            */
/* --------------------------------------------------------------- */
// This could be merged into BitSetIO as another action, but as it
// will be used rarely, perhaps, is better if omittable
int bitExistIO(uint port, uint mask) {
  switch(port) {
    case 0:
      // 14:12 not available
      // 29:29 are shared with USB D- and D+ and are specially
      //       treated, so are also shown as not available here
      if ((0xCFFF8FFF & mask) == mask) return 1;
      break;

    case 1:
      // 2:3, 7:5, 11:13 not available
      if ((0xFFFFC713 & mask) == mask) return 1;
      break;

    case 2:
      // 14:31 not available
      if ((0x00003FFF & mask) == mask) return 1;
      break;

    case 3:
      // only 26:25 available
      if ((0x06000000 & mask) == mask) return 1;
      break;

    case 4:
      // only 29:28 available
      if ((0x30000000 & mask) == mask) return 1;
      break;

    default:        // bad port #
      break;
    } // switch

  return 0;
  } // bitExistIO


/* --------------------------------------------------------------- */
/* bitHandler -- master handler for bit interrupts                 */
/*                                                                 */
/* All interrupts for bits arrive here; the hendler determines     */
/* which bit caused the interrupt and then calls the individual    */
/* bit handler.                                                    */
/*                                                                 */
/* The procedure is to read the four port status registers at the  */
/* start, and then process by port.  For each pin for which either */
/* a rising or falling edge interrupt has been generated, call the */
/* pin's interrupt handler (that handler can read the bit to find  */
/* out which happened, if BIT_RISEFALL is in use).                 */
/*                                                                 */
/* Once all the handlers have been called for a port the           */
/* interrupts that were noted are cleared.                         */
/* --------------------------------------------------------------- */
static void bitHandler(void) {
  // note pending interrupts
  uint pend0=LPC_GPIOINT->IO0IntStatR | LPC_GPIOINT->IO0IntStatF;
  uint pend2=LPC_GPIOINT->IO2IntStatR | LPC_GPIOINT->IO2IntStatF;
  int  i;                               // work
  uint bit;                             // ..

  if (pend0) {                          // at least one PORT0 pending
    bit=0x01;                           // bit to test
    for (i=0; i<BITS0; i++) {
      if (pend0 & bit) {
        // if there's a handler set for the bit, invoke it
        // [ignore spurious interrupts, possibly set up 'manually']

        if (bithandlers0[i]) bithandlers0[i]();

        }
      bit<<=1;
      } // i
    LPC_GPIOINT->IO0IntClr=pend0;       // clear
    } // P0

  if (pend2) {                          // at least one PORT2 pending
    bit=0x01;                           // bit to test
    for (i=0; i<BITS2; i++) {
      if (pend2 & bit) {
        // if there's a handler set for the bit, invoke it
        // [ignore spurious interrupts, possibly set up 'manually']
        if (bithandlers2[i]) bithandlers2[i]();
        }
      bit<<=1;
      } // i
    LPC_GPIOINT->IO2IntClr=pend2;       // clear
    } // P2
  } // bitHandler

/* --------------------------------------------------------------- */
/* bitIntIO -- effect an interrupt configuration of a bit or bits  */
/*                                                                 */
/*   port    -- port in which the bit or bits reside               */
/*   mask    -- mask of which bit(s) to affect                     */
/*   action  -- one of BIT_RISE, _FALL, _RISEFALL, or _DISABLE     */
/*   handler -- interrupt handler to call, or NULL if _DISABLE     */
/*                                                                 */
/*   returns 0 if sucessful or negative if error (e.g., the bit    */
/*     does not exist or does not support interrupts)              */
/*                                                                 */
/* This first disables all interrupts for the specified bit(s).    */
/* It then enables the pins as specified, unless handler is NULL.  */
/*                                                                 */
/* The bit validity check only tests for whether all the specified */
/* bits are valid for interrupts, not whether they are brought out */
/* to a physical pin (they could still be used for self-wrought    */
/* interrupts).                                                    */
/* --------------------------------------------------------------- */
int bitIntIO(uint port, uint mask, uint action, vector handler) {
  vector *harray=NULL;             // which handler array to use
  int  i, rc;                      // work
  uint bit;                        // ..

  // check valid port, bits, etc., and note which array to use
  // disable and clear rise & fall interrupts for the requested bits,
  // then set the matching bithandler addresses

  if (port==PORT0) {
    if ((mask&0x80007000)!=0) return -1;     // check port0 bits
    LPC_GPIOINT->IO0IntEnR &= ~mask;         // clear rising interrupts
    LPC_GPIOINT->IO0IntEnF &= ~mask;         // clear falling interrupts
    harray=bithandlers0;                     // array to use
    }
   else if (port==PORT2) {
    if ((mask&0xFFFFC000)!=0) return -2;     // check port2 bits
    LPC_GPIOINT->IO2IntEnR &= ~mask;         // clear rising interrupts
    LPC_GPIOINT->IO2IntEnF &= ~mask;         // clear falling interrupts
    harray=bithandlers2;                     // array to use
    }
   else return -3;                           // bad port request

  bit=0x01;                                  // bit to test
  for (i=0; i<32; i++) {
    if (mask & bit) harray[i]=handler;
    bit<<=1;
    } // i
  if (action==BIT_DISABLE) return 0;         // disable request complete

  // set master interrupt handler if not yet set up; this will also
  // enable the interrupt
  // [error if already in use by a different handler]
  rc=sysVector(sysEnumVector(EINT3_IRQn), bitHandler);
  if (rc<0) return -4;                       // already in use

  // finally set interrupt requests for the bits
  // (Cannot use GPIO_IntCmd because it overwrites)
  if (action==BIT_RISE || action==BIT_RISEFALL) {
    if (port==PORT0) LPC_GPIOINT->IO0IntEnR |= mask;
     else            LPC_GPIOINT->IO2IntEnR |= mask;
    }
  if (action==BIT_FALL || action==BIT_RISEFALL) {
    if (port==PORT0) LPC_GPIOINT->IO0IntEnF |= mask;
     else            LPC_GPIOINT->IO2IntEnF |= mask;
    }
  return 0;
  } // bitIntIO

/* --------------------------------------------------------------- */
/* bitQueryIO -- query the state of a bit or bits                  */
/*                                                                 */
/*   port   -- port in which the bit or bits reside                */
/*   mask   -- mask of which bit(s) to query                       */
/*   bits   -- receives the state of the selected bits (other      */
/*             bits will be 0)                                     */
/*                                                                 */
/*   returns 0 if OK, -1 if error                                  */
/*                                                                 */
/* The result will be -1 if port is not a valid port number.       */
/* Bits are undefined if the selected bits do not exist.           */
/* Bits may also be undefined or unpredictable if the pin          */
/* corresponding to a bit is in alternative use (e.g., for ADC).   */
/* --------------------------------------------------------------- */
int bitQueryIO(uint port, uint mask, uint *bits) {
  if (port>2) return -1;
  *bits=GPIO_ReadValue(port) & mask;
  return 0;
  } // bitQueryIO

/* --------------------------------------------------------------- */
/* bitSetIO -- effect an action on a bit or bits                   */
/*                                                                 */
/*   port   -- port in which the bit or bits reside                */
/*   mask   -- mask of which bit(s) to affect                      */
/*   action -- one of BIT_ON, BIT_OFF, or BIT_INVERT for bit       */
/*             output, or one of BIT_PULLUP, BIT_TRISTATE, or      */
/*             BIT_PULLDOWN for bit input                          */
/*                                                                 */
/*   returns 0 if OK, -1 if error                                  */
/*                                                                 */
/* BIT_ON, BIT_OFF, or BIT_INVERT set the specified bit(s) to      */
/* output mode then sets the bit(s) as appropriate for the request */
/* (BIT_OFF sets each bit to 0, BIT_ON sets them to 1, BIT_INVERT  */
/* flips each bit).                                                */
/*                                                                 */
/* The actions BIT_PULLUP, BIT_TRISTATE, and BIT_PULLDOWN set the  */
/* specified bit(s) to input mode with pull-up, neither pull-up    */
/* nor pull-down, or pull-down.                                    */
/*                                                                 */
/* After reset, bits are BIT_TRISTATE.                             */
/*                                                                 */
/* For example:                                                    */
/*                                                                 */
/*   rc=bitSet(PORT1, BIT27, BIT_PULLUP);                          */
/*                                                                 */
/* The result is -1 if port is not a valid port number.            */
/* Bits are undefined if the selected bits do not exist.           */
/* --------------------------------------------------------------- */
int bitSetIO(uint port, uint mask, uint action) {

  if (port>2) return -1;

  // set direction
  GPIO_SetDir(port, mask, (action&BIT_OUTPUT)!=0);

  switch(action) {
    case BIT_OFF:
      GPIO_ClearValue(port, mask);
      break;

    case BIT_ON:
      GPIO_SetValue(port, mask);
      break;

    case BIT_INVERT: {
      uint read=GPIO_ReadValue(port);
      GPIO_ClearValue(port, mask&read);      // clear those that were 1
      GPIO_SetValue(port, mask&(~read));     // set those that were 0
      break;}

    case BIT_PULLUP:
    case BIT_PULLDOWN:
    case BIT_TRISTATE: {
      PINSEL_CFG_Type pincfg;                // for configuring pins
      uint bit;                              // work
      int i;                                 // ..

      // Set up common configuration
      pincfg.Funcnum=PINSEL_FUNC_0;          // pins are all bank 0
      pincfg.Portnum=port;
      pincfg.OpenDrain=PINSEL_PINMODE_NORMAL;

      if (action==BIT_PULLUP) pincfg.Pinmode=PINSEL_PINMODE_PULLUP;
       else if (action==BIT_PULLDOWN) pincfg.Pinmode=PINSEL_PINMODE_PULLDOWN;
       else pincfg.Pinmode=PINSEL_PINMODE_TRISTATE;

      // now configure all requested pins
      bit=0x01;
      for (i=0; i<32; i++) {
        if (mask & bit) {
          pincfg.Pinnum=i;
          PINSEL_ConfigPin(&pincfg);
          }
        bit<<=1;
        } // i
      break;}

    default:        // unknown action
      return -2;
      break;
    } // switch

  return 0;
  } // bitSetIO


