/* --------------------------------------------------------------- */
/* lpcSysIO.c -- system-dependent core functions                   */
/* --------------------------------------------------------------- */
/*                                                                 */
/* Copyright (c) Mike Cowlishaw, 2010.  All rights reserved.       */
/*                                                                 */
/* Permission to use, copy, modify, and/or distribute this         */
/* software for any purpose with or without fee is hereby granted, */
/* provided that the above copyright notice and this permission    */
/* notice appear in all copies, and that notice and the date of    */
/* any modifications be added to the software.                     */
/*                                                                 */
/* This software is provided "as is".  No warranties, whether      */
/* express, implied, or statutory, including, but not limited to,  */
/* implied warranties of merchantability and fitness for a         */
/* particular purpose apply to this software.  The author shall    */
/* not, in any circumstances, be liable for special, incidental,   */
/* or consequential damages, for any reason whatsoever.            */
/*                                                                 */
/* --------------------------------------------------------------- */
/* These are processor-specific functions for Tollos.              */
/*                                                                 */
/* They should not normally be used directly by applications.      */
/* --------------------------------------------------------------- */

#define  TOLLOS_IO                 // allow use of I/O functions
#include "tollos.h"
#include "LPC17xx.h"
#include "lpc17xx_clkpwr.h"        // for power settings
#include "lpc17xx_systick.h"       // for SysTick
#include "lpc17xx_wdt.h"           // for watchdog
#include "system_LPC17xx.h"        // for SystemInit

// Local statics
static int  msSysTick=0;      // SysTick interval; 0 until first set
static uint watchtime=0;      // Watchdog ms; 0 until first set
static void stopPLL0(void);   // turn off PLL0 as per errata

/* --------------------------------------------------------------- */
/* sysFeedDogIO -- feed the watchdog (if started)                  */
/*                                                                 */
/* Once the watchdog has been started, this function is called to  */
/* reset its counter and prevent a Watchdog Reset.  If called      */
/* before the watchdog is started it has no effect.                */
/* --------------------------------------------------------------- */
void sysFeedDogIO(void) {
  if (watchtime!=0) WDT_Feed();
  } // sysFeedDogIO

/* --------------------------------------------------------------- */
/* sysInitProcIO -- initialize processor base function [startup]   */
/*                                                                 */
/*   returns 0 if OK, -1 if error (but probably cannot recover)    */
/*                                                                 */
/* This must be called before any other sysIO functions; it:       */
/*   -- initializes the processor [SystemInit]                     */
/*   -- sets the system core clock frequency (processor speed)     */
/*   -- maps the vector table to the start of RAM                  */
/*                                                                 */
/* --------------------------------------------------------------- */
// Define required core clock speed
#define CLOCKMHZ  96          // must be a multiple of 6 (see below)
int sysInitProcIO(void) {     // processor initialize (clocks, etc.)

  /* ------------------------------------------------------------- */
  /* Processor initialization (clock setup, etc.)                  */
  /*                                                               */
  /* This is called after C initialization and the vector table    */
  /* has been copied to RAM.                                       */
  /* ------------------------------------------------------------- */

  SystemInit();               // use the required CMSIS function
  // this set the clock to some default, e.g., 64MHz; now change
  // it to the desired frequency; divider has been set to 4

  // setPLL0 frequency; the constraints are:
  //   Fin must be in range 32 kHz to 50 MHz [fixed at 12 MHz]
  //   Fcco = (2 * M * 12000000 / N) must be 275 MHz -> 550 MHz
  // the output clock will be Fcco/4  (the 4 could be changed)
  //   = M * 6 MHz if N=1
  #define M  (CLOCKMHZ/6)     // 16 for 96 MHz, 12 for 72 MHz
  #define N  1

  LPC_SC->CLKSRCSEL = 1;                     // main oscillator
  LPC_SC->PLL0CFG   = (((unsigned int)N-1) << 16) | (M-1);
  LPC_SC->PLL0FEED  = 0xAA;
  LPC_SC->PLL0FEED  = 0x55;
  LPC_SC->PLL0CON   = 0x01;                  // enable
  LPC_SC->PLL0FEED  = 0xAA;
  LPC_SC->PLL0FEED  = 0x55;
  // wait for PLL lock (PLOCK0, no equate supplied)
  while (!(LPC_SC->PLL0STAT & (1<<26)));

  LPC_SC->PLL0CON   = 0x03;                  // enable & connect
  LPC_SC->PLL0FEED  = 0xAA;
  LPC_SC->PLL0FEED  = 0x55;
  // Wait for PLLC0_STAT & PLLE0_STAT (unequated)
  // [NB: NXP code appears to be wrong for this]
  #define STATBITS ((1<<25) | (1<<24))
  while ((LPC_SC->PLL0STAT & STATBITS) != STATBITS);

  SystemCoreClockUpdate();                   // required (see LPC docs)

  /* ------------------------------------------------------------- */
  /* set vector table location to RAM start                        */
  /* ------------------------------------------------------------- */
  SCB->VTOR=(uint32_t)&chipRAMbase & 0x3FFFFF80;

  return 0;
  } // sysInitProcIO

/* --------------------------------------------------------------- */
/* sysIRQIO -- enable or disable an interrupt                      */
/*                                                                 */
/*   index  -- vector table index number (see sysVector)           */
/*   enable -- 1 to enable, 0 to disable                           */
/*                                                                 */
/*   returns 0 if OK, 1 if 0<index<16 (system vector), -1 if index */
/*     is too large                                                */
/* --------------------------------------------------------------- */
int  sysIRQIO(uint index, flag enable) {

  if (index<16) return 1;
  if (index>=(int)&chipVectors) return -1;

  if (enable) NVIC_EnableIRQ(index-16);
   else NVIC_DisableIRQ(index-16);
  return 0;
  } // sysIRQIO

/* --------------------------------------------------------------- */
/* sysPowerIO -- set a power mode and wait for interrupt           */
/*                                                                 */
/*   item -- a constant listed in tollosIO.h                       */
/*                                                                 */
/* Valid items are:                                                */
/*   SYSTEM_DEEPPOWERDOWN -- power off until Reset or RTC wake     */
/*   SYSTEM_DEEPSLEEP     -- unclocked sleep                       */
/*   SYSTEM_POWERDOWN     -- power down (unclocked, slow start)    */
/*   SYSTEM_SLEEP         -- idle (Wait For Interrupt)             */
/*                                                                 */
/* Currently Tollos only uses SLEEP and DEEPPOWERDOWN.             */
/* --------------------------------------------------------------- */
void sysPowerIO(uint item)  {
  switch(item) {
    case SYSTEM_SLEEP: {       // Sleep (idle)
      CLKPWR_Sleep();
      break;}
    case SYSTEM_DEEPSLEEP: {   // Sleep (unclocked)
      stopPLL0();
      CLKPWR_DeepSleep();
      break;}
    case SYSTEM_POWERDOWN: {   // Power down (unclocked, slow start)
      stopPLL0();
      CLKPWR_PowerDown();
      break;}
    case SYSTEM_DEEPPOWERDOWN:{// Power off until Reset or RTC
      stopPLL0();
      CLKPWR_DeepPowerDown();
      break;}
    default: {                // ignore; return immediately
      break;}
    } // switch
  } // sysPowerIO

/* --------------------------------------------------------------- */
/* sysQueryIO -- query a system item                               */
/*                                                                 */
/*   item -- a constant listed in tollosIO.h                       */
/*                                                                 */
/*   returns current value of the item; negative generally         */
/*     indicates failure                                           */
/*                                                                 */
/* Valid items are:                                                */
/*   SYSTEM_CLOCK    -- clock speed, Hz                            */
/*   SYSTEM_RESET    -- 0 if last reset was normal (POR or Reset   */
/*                      pin), 1 if abnormal (Watchdog or Brownout) */
/*   SYSTEM_SYSTICK  -- SysTick interval in ms                     */
/*   SYSTEM_WATCHDOG -- Watchdog interval in ms                    */
/*                                                                 */
/* --------------------------------------------------------------- */
int sysQueryIO(uint item)  {
  int rc;                     // value to return
  switch(item) {

    case SYSTEM_CLOCK: {      // clock speed, Hz
      rc=SystemCoreClock;     // from the horse's mouth
      break;}

    case SYSTEM_RESET: {      // 0 = normal, 1 = abnormal
      // Watchdog WDTR=bit2 or brownout BODR=bit3 have no equates
      rc=(LPC_SC->RSID & 0x0C)!=0;
      break;}

    case SYSTEM_SYSTICK: {    // current SysTick interval
      rc=msSysTick;
      break;}

    case SYSTEM_WATCHDOG: {   // current Watchdog timeout
      rc=watchtime;
      break;}

    default: {
      rc=-1;                  // bad item
      break;}
    } // switch
  return rc;
  } // sysQueryIO

/* --------------------------------------------------------------- */
/* sysSetIO -- set a system item                                   */
/*                                                                 */
/*   item  -- a constant listed in tollosIO.h                      */
/*   value -- value to set the item to                             */
/*                                                                 */
/*   returns previous value of the item, or a negative number      */
/*     if an error or unknown                                      */
/*                                                                 */
/* Valid items are:                                                */
/*   SYSTEM_SYSTICK  -- SysTick interval in ms [>0, maximum        */
/*                      depends on clock rate]                     */
/*   SYSTEM_WATCHDOG -- Watchdog interval in ms [0<value<4000000]  */
/*                                                                 */
/* --------------------------------------------------------------- */
int sysSetIO(uint item, int value)  {
  int rc;                // value to return
  switch(item) {

    case SYSTEM_SYSTICK: {              // set SysTick [ms]
      uint count;                       // work
      rc=msSysTick;

      // check parameter
      if (value<=0) return -1;          // must be at least 1

      // calculate SysTick counter (number to divide clock by); this
      // must be no more than 0xFFFFFF; calculate carefully to
      // avoid multiplication overflow
      count=SystemCoreClock/1000 * value;
      if (count>0xFFFFFF) return -1;    // SysTick counter out of range

      // The CMSSIS quick setter
      if (SysTick_Config(count)) rc=-1; // failed
       else msSysTick=value;
      break;}

    case SYSTEM_WATCHDOG: {             // watchdog timeout
      uint ms=(uint)value;
      if (ms>4000000 || ms<1) return -1;
      rc=watchtime;
      if (watchtime==0) {     // first call
        WDT_Init(WDT_CLKSRC_IRC, WDT_MODE_RESET);
        WDT_Start(value*1000U);
        }
       else WDT_UpdateTimeOut(value*1000U);  // [WDT_Start less Enable]
      watchtime=value;
      break;}

    default: {
      rc=-1;             // bad item
      break;}
    } // switch
  return rc;
  } // sysSetIO


/* --------------------------------------------------------------- */
/* stopPLL0 -- turn off PLL                                        */
/* --------------------------------------------------------------- */
static void stopPLL0(void) {  // turn off PLL0 as per errata
  // needed to stop it continuing to use power
  LPC_SC->PLL0CON &= ~(1<<1); // disconnect PLL0
  LPC_SC->PLL0FEED = 0xAA;    // feed
  LPC_SC->PLL0FEED = 0x55;    // ..
  while ((LPC_SC->PLL0STAT & (1<<25)) != 0x00); // Wait for PLL0 to disconnect
  LPC_SC->PLL0CON &= ~(1<<0); // turn off PLL0
  LPC_SC->PLL0FEED = 0xAA;    // feed
  LPC_SC->PLL0FEED = 0x55;    // ..
  while ((LPC_SC->PLL0STAT & (1<<24)) != 0x00); // Wait for PLL0 to shut down
  } // stopPLL0

