/* --------------------------------------------------------------- */
/* lpcUARTIO.c -- UART device I/O functions                        */
/* --------------------------------------------------------------- */
/*                                                                 */
/* Copyright (c) Mike Cowlishaw, 2010.  All rights reserved.       */
/*                                                                 */
/* Permission to use, copy, modify, and/or distribute this         */
/* software for any purpose with or without fee is hereby granted, */
/* provided that the above copyright notice and this permission    */
/* notice appear in all copies, and that notice and the date of    */
/* any modifications be added to the software.                     */
/*                                                                 */
/* This software is provided "as is".  No warranties, whether      */
/* express, implied, or statutory, including, but not limited to,  */
/* implied warranties of merchantability and fitness for a         */
/* particular purpose apply to this software.  The author shall    */
/* not, in any circumstances, be liable for special, incidental,   */
/* or consequential damages, for any reason whatsoever.            */
/*                                                                 */
/* --------------------------------------------------------------- */
/* The UART functions provide basic I/O over a serial link.        */
/*                                                                 */
/* --------------------------------------------------------------- */
// not fully tested yet for all UARTs

#define UART_DEVS 4           // largest UART number + 1

#define  TOLLOS_IO            // allow use of I/O functions
#include "tollos.h"
#include "lpc17xx_uart.h"     // Central include file
#include "lpc17xx_pinsel.h"

// statics
static LPC_UART_TypeDef * devs[UART_DEVS]=   // UART devices
       {LPC_UART0,
       (LPC_UART_TypeDef *)LPC_UART1,        // UART1 is LPC special
       LPC_UART2,
       LPC_UART3};   //   ..

/* --------------------------------------------------------------- */
/* uartGetcIO -- get character from UART device                    */
/*                                                                 */
/*   uart -- ID of UART to read                                    */
/*                                                                 */
/*   returns character read from UART, or -1 if none there or UART */
/*     does not exist                                              */
/*                                                                 */
/* This function does not block.  It returns an int so that errors */
/* are distinct, and will also allow extension.                    */
/* --------------------------------------------------------------- */
int uartGetcIO(uint uart) {
  LPC_UART_TypeDef *dev;

  if (uart>=UART_DEVS) return -1;
  dev=devs[uart];

  if (!(dev->LSR & UART_LSR_RDR)) return -1; // nothing pending

  return UART_ReceiveByte(dev);              // uint8_t promoted
  } // uartGetcIO

/* --------------------------------------------------------------- */
/* uartInitIO -- initialize port for I/O                           */
/*                                                                 */
/*   uart -- ID of UART to initialize                              */
/*   baud -- rate to run uart communications                       */
/*                                                                 */
/*   returns 0 if OK, or -1 if uart does not exist                 */
/* --------------------------------------------------------------- */
int uartInitIO(uint uart, uint baud) {
  UART_CFG_Type   uartcfg;         // UART configuration
  PINSEL_CFG_Type pincfg;          // pin configuration
  uint txpin, rxpin;               // work

  // initialize UART pin configuratins
  pincfg.Portnum=PORT0;                      // all UARTs are on port 0
  pincfg.OpenDrain=PINSEL_PINMODE_NORMAL;    // modes all the same ..
  pincfg.Pinmode=PINSEL_PINMODE_PULLUP;      // ..
  pincfg.Funcnum=PINSEL_FUNC_1;              // most UARTs are in bank 1
  // now the UART-specific settings
  switch(uart) {
    case 0:
      txpin=2;
      rxpin=3;
      break;
    case 1:
      txpin=15;
      rxpin=16;
      break;
    case 2:
      txpin=10;
      rxpin=11;
      break;
    case 3:
      txpin=0;
      rxpin=1;
      pincfg.Funcnum=PINSEL_FUNC_2;          // this one is bank 2
      break;
    default: return -1;
    } // switch

  pincfg.Pinnum=txpin;
  PINSEL_ConfigPin(&pincfg);
  pincfg.Pinnum=rxpin;
  PINSEL_ConfigPin(&pincfg);

  // initialize uartcfg to default:
  //   baudrate = 9600bps, 8 data bits, 1 stop bit, no parity
  UART_ConfigStructInit(&uartcfg);
  // [no error returned]

  // set baud rate field as requested
  uartcfg.Baud_rate=baud;

  // initialize UART [no error returned]
  UART_Init(devs[uart], &uartcfg);

  // enable UART transmit [no error returned]
  UART_TxCmd(devs[uart], ENABLE);

  return 0;
  } // uartInitIO

/* --------------------------------------------------------------- */
/* uartPutcIO -- send character to UART device                     */
/*                                                                 */
/*   uart -- ID of UART to send character to                       */
/*   ch   -- character to send                                     */
/*                                                                 */
/*   returns 0 if OK, or -1 if uart does not exist                 */
/*                                                                 */
/* This waits for any previous send to complete.                   */
/* The character, ch, is an int to allow future extension.         */
/* --------------------------------------------------------------- */
int uartPutcIO(uint uart, int ch) { // send character
  unsigned char bi[2];
  if (uart>=UART_DEVS) return -1;
  bi[0]=ch;
  bi[1]='\0';
  UART_Send(devs[uart], bi, 1, BLOCKING);  // will always return 1
  return 0;
  } // uartPutcIO

