/* --------------------------------------------------------------- */
/* mbedLED.c -- LED device I/O functions                           */
/* --------------------------------------------------------------- */
/*                                                                 */
/* Copyright (c) Mike Cowlishaw, 2010.  All rights reserved.       */
/*                                                                 */
/* Permission to use, copy, modify, and/or distribute this         */
/* software for any purpose with or without fee is hereby granted, */
/* provided that the above copyright notice and this permission    */
/* notice appear in all copies, and that notice and the date of    */
/* any modifications be added to the software.                     */
/*                                                                 */
/* This software is provided "as is".  No warranties, whether      */
/* express, implied, or statutory, including, but not limited to,  */
/* implied warranties of merchantability and fitness for a         */
/* particular purpose apply to this software.  The author shall    */
/* not, in any circumstances, be liable for special, incidental,   */
/* or consequential damages, for any reason whatsoever.            */
/*                                                                 */
/* --------------------------------------------------------------- */
/* A function to effect an action on, or query,  an LED (assumed   */
/* to be board-mounted).                                           */
/*                                                                 */
/* LED numbering is arbitrary, but if there is one or more board   */
/* LEDs then one should be accessible as 'LED 1'.                  */
/*                                                                 */
/* If there is only the one LED on the board and documentation     */
/* suggests that it be called 'LED 0' then it is recommended that  */
/* both num=0 and num=1 map to the same LED.                       */
/* --------------------------------------------------------------- */


#define  TOLLOS_IO            // allow use of I/O functions
#include "tollos.h"
#include "tollosLED.h"        // to check signatures
#include "LPC17xx.h"
#include "lpc17xx_gpio.h"

/* --------------------------------------------------------------- */
/* ledIO --  effect an I/O action on a LED                         */
/*                                                                 */
/*   num    -- number of the LED (board-defined)                   */
/*   action -- one of LED_ON, _OFF, _INVERT, _QUERY                */
/*                                                                 */
/*   returns the previous value of the LED (0 or 1), or -1 if num  */
/*     or action is invalid                                        */
/* --------------------------------------------------------------- */
int ledIO(int num, uint action) {       // carry out LED I/O action
  static int ready=0;                   // 1 once set up
  uint was;                             // previous setting
  uint bit, mask;                       // bit and corresponding mask

  switch (num) {
    case 1: bit=18; break;
    case 2: bit=20; break;
    case 3: bit=21; break;
    case 4: bit=23; break;
    default:
      return -1;                        // not 1 through 4
    }
  mask=1<<bit;

  if (!ready) {                         // configure needed
    ready=1;
    GPIO_SetDir(1, 0x00B40000, 1);      // 4 bits in Port1 to output
    }

  was=(GPIO_ReadValue(1)&mask)!=0;      // current value

  switch (action) {
    case LED_ON:
      GPIO_SetValue(1, mask);
      break;
    case LED_OFF:
      GPIO_ClearValue(1, mask);
      break;
    case LED_INVERT:
      if (was) GPIO_ClearValue(1, mask);
        else GPIO_SetValue(1, mask);
      break;
    case LED_QUERY:
      // no change needed
      break;
    default:
      return -1;
    }

  return was;
  } // ledIO


