/* --------------------------------------------------------------- */
/* tollosDebug.c -- debug functions for Tollos                     */
/* --------------------------------------------------------------- */
/*                                                                 */
/* Copyright (c) Mike Cowlishaw, 2010.  All rights reserved.       */
/*                                                                 */
/* Permission to use, copy, modify, and/or distribute this         */
/* software for any purpose with or without fee is hereby granted, */
/* provided that the above copyright notice and this permission    */
/* notice appear in all copies, and that notice and the date of    */
/* any modifications be added to the software.                     */
/*                                                                 */
/* This software is provided "as is".  No warranties, whether      */
/* express, implied, or statutory, including, but not limited to,  */
/* implied warranties of merchantability and fitness for a         */
/* particular purpose apply to this software.  The author shall    */
/* not, in any circumstances, be liable for special, incidental,   */
/* or consequential damages, for any reason whatsoever.            */
/*                                                                 */
/* --------------------------------------------------------------- */
/* The debug device uses a serial channel to communicate with a    */
/* terminal device (usually a termulator emulator application on   */
/* a PC).                                                          */
/*                                                                 */
/* Using any debug function automatically initializes the debug    */
/* communication device.                                           */
/* The default baud rate for the device is TOLLOS_DEBUG_BAUD,      */
/* normally 115200 baud.                                           */
/* --------------------------------------------------------------- */

#define  TOLLOS_IO            // allow use of I/O functions
#include "tollosDebug.h"
#include <stdarg.h>
#include <string.h>

static flag initialized=0;    // 1 if initialized

// ---------------------------------------------------------------
// putc -- proxy to keep debugf code more portable & readable
// ---------------------------------------------------------------
#define putc debugPutcIO      // just a name change

/* --------------------------------------------------------------- */
/* debugBaud -- set debug baud rate                                */
/*                                                                 */
/*   baud -- baud rate to use for debug                            */
/*                                                                 */
/* It is recommend that a 'standard' baud rate be used, for        */
/* example, 9600 or 115200.                                        */
/* At 9600 baud about 38 characters can be sent in 40ms (one       */
/* default tick) at 115200, 460 characters can be sent.            */
/* --------------------------------------------------------------- */
void debugBaud(uint baud) {   // initialize
  debugInitIO(baud);
  sysPause(0);                // dummy link
  initialized=1;
  } // debugBaud

/* --------------------------------------------------------------- */
/* debugf -- a mini printf to send a string                        */
/*                                                                 */
/*   format -- format and data string                              */
/*   ...    -- other arguments as required                         */
/*                                                                 */
/*   returns the number (count) of characters sent                 */
/*                                                                 */
/* The following format indicators are supported:                  */
/*                                                                 */
/*   %%   -- '%'                                                   */
/*   %c   -- single char                                           */
/*   %d,i -- int                                                   */
/*   %m   -- minidec decimal number  [iff MINIDEC is defined]      */
/*   %s   -- string                                                */
/*   %u   -- unsigned int                                          */
/*   %x   -- hexadecimal                                           */
/*   %\n  -- LF only (otherwise ""\n"" becomes CRLF)               */
/*                                                                 */
/* The four ""printf"" format flags (""+ - 0 #"") are allowed but  */
/* are ignored, except:                                            */
/*   '+' on integers and decimals (requests + signs)               */
/*   '0' on 'x' (e.g., ""%08x"" for an 8-hexadecimal-digit result) */
/*   '#' on 'x' (prefixes '0x')                                    */
/*                                                                 */
/* The 'h', 'l', and 'L' modifiers are not supported.              */
/*                                                                 */
/* Field width is ignored except for 'x' (where it is capped to 8) */
/* and decimals ('m'); the latter will fill with "****" if the     */
/* number does not fit in the specified field width.               */
/*                                                                 */
/* Field precision (e.g., %5.7) is not supported; for decimals,    */
/* use Quantize on a number to set its 'precision'.                */
/*                                                                 */
/* A ""\n"" in the string is expanded to ""\r\n"" unless preceded  */
/* by ""%"".                                                       */
/* Almost no error checking is carried out.                        */
/* --------------------------------------------------------------- */
int debugf(const char *format, ...) {
  va_list args;          // argument list
  const char *f;         // pointer in format string
  int count=0;           // chars put counter
  uint width=0;          // field width [0 means none]
  flag sign;             // 1 if '+' flag
  flag neg;              // 1 if '-' flag
  flag zero;             // 1 if '0' flag
  flag alt;              // 1 if '#' flag

  if (!initialized) debugBaud(TOLLOS_DEBUG_BAUD);

  va_start(args, format);

  for (f=format; *f!=0; f++) {
    if (*f!='%') {
      if (*f=='\n') {putc('\r'); count++;}
      putc(*f); count++;
      continue;
      }
    // have %
    f++;
    if (*f=='%') {putc(*f); count++; continue;}

    neg=0; sign=0; alt=0; zero=0;
    for (;; f++) {           // handle leading flags in any order
      if (*f=='-') neg=1;
      else if (*f=='+') sign=1;
      else if (*f=='0') zero=1;    // (strips leading zeros)
      else if (*f=='#') alt=1;
      else break;
      } // flags loop
    width=0;                       // gather width, if any
    for (; *f>='0' && *f<='9'; f++) width=width*10+(*f-'0');
    // [dot and precision would be gathered here]

    switch(*f) {                   // format indicator
      case 'c': {
        char c=(char)va_arg(args, int);
        putc(c); count++;
        break;}

      case 'u':
      case 'd':
      case 'i': {
        uint u=(uint)va_arg(args, uint);
        uint p;          // powers of 10
        flag had;        // had leading digit
        if (*f!='u') {   // signed
          int i=(int)u;
          if (i<0) {
            putc('-'); count++;
            if (u==0x80000000) u=2147483648U;
             else u=-i;
            }
           else if (sign) {putc('+'); count++;}
          }
         else if (sign) {putc('+'); count++;}
        // u is now abs(arg), such that 0 <= u <= 4294967295
        had=0;
        if (u==0)  {putc('0'); count++;}
         else for (p=1000000000; p; p/=10) {
          uint d=u/p;
          if (d) had=1;  // had leading digit
          u%=p;
          if (had) {putc('0'+d); count++;}
          }
        break;}

      #ifdef MINIDEC
      case 'm': {
        dec num=(dec)va_arg(args, dec);

        char buff[20];   // work, >17
        char *cp=buff;   // work

        if (sign && !decIsNegative(num)) {buff[0]='+'; cp++;}
        decToString(num, cp);

        if (width!=0) {
          int len=strlen(buff);
          int pad=width-len;
          if (pad<0) {
            for (; width>0; width--) {putc('*'); count++;} // won't fit
            break;}
          for (; pad>0; pad--) {putc(' '); count++;}
          }
        for (cp=buff; *cp; cp++) {putc(*cp); count++;}
        break;}
      #endif

      case 's': {
        char *cp=(char *)va_arg(args, int);
        if (cp!=NULL) for (; *cp; cp++) {putc(*cp); count++;}
        break;}

      case 'x':
      case 'X': {
        uint u=(uint)va_arg(args, uint);
        byte upper=(*f=='X');
        flag had;
        char x[8];
        int i;
        // convert to hex
        for (i=6; i>=0; i-=2) {
          byte2hex(u & 0xff, &x[i], upper);
          u>>=8;
          }
        if (width==0 || width>8) width=8;
        if (alt) {putc('0'); count++; putc(*f); count++;}
        had=0;
        for (i=8-width; i<8; i++) {
          if (!zero && !had && x[i]=='0') continue;    // skip leading 0s
          had=1;
          putc(x[i]); count++;
          }
        break;}


      default: {
        putc(*f); count++;
        break;}
      } // format char

    } // format scan

  va_end(args);     // done with argument list
  return count;
  } // debugf


/* --------------------------------------------------------------- */
/* debugGet -- get character from host                             */
/*                                                                 */
/*   returns a character read, or -1 if none pending               */
/*                                                                 */
/* If a character was read the value returned will always be       */
/* non-negative (hence all values >=0 might be possible).          */
/*                                                                 */
/* Note that some keyboard keys may generate more than one         */
/* character: an escape followed by some code, for example.        */
/* --------------------------------------------------------------- */
int debugGet(void) {          // get character
  if (!initialized) debugBaud(TOLLOS_DEBUG_BAUD);
  return debugGetcIO();
  } // debugGet

