/* --------------------------------------------------------------- */
/* tollosUtil.c -- useful macros and functions for Tollos          */
/* --------------------------------------------------------------- */
/*                                                                 */
/* Copyright (c) Mike Cowlishaw, 2010.  All rights reserved.       */
/*                                                                 */
/* Permission to use, copy, modify, and/or distribute this         */
/* software for any purpose with or without fee is hereby granted, */
/* provided that the above copyright notice and this permission    */
/* notice appear in all copies, and that notice and the date of    */
/* any modifications be added to the software.                     */
/*                                                                 */
/* This software is provided "as is".  No warranties, whether      */
/* express, implied, or statutory, including, but not limited to,  */
/* implied warranties of merchantability and fitness for a         */
/* particular purpose apply to this software.  The author shall    */
/* not, in any circumstances, be liable for special, incidental,   */
/* or consequential damages, for any reason whatsoever.            */
/*                                                                 */
/* --------------------------------------------------------------- */
/*                                                                 */
/* The macros are defined in ""tollosUtil.h"", and are:            */
/*                                                                 */
/*   abs(x)     -- return the absolute value of x                  */
/*   iswap(x,y) -- swap the values on the ints x and y             */
/*   max(a,b)   -- return the larger of x and y                    */
/*   min(a,b)   -- return the smaller of x and y                   */
/*   signum(x)  -- return the sign of x (-1, 0, or 1)              */
/*                                                                 */
/* The functions are miscellaneous.                                */
/* --------------------------------------------------------------- */

#include "tollos.h"

/* --------------------------------------------------------------- */
/* byte2hex -- convert a byte to two hex chars                     */
/*                                                                 */
/*   b       -- byte to convert                                    */
/*   hex     -- points to where ASCII bytes should go              */
/*   upper   -- 1 to use ABCDEF, 0 for abcdef                      */
/*                                                                 */
/* This writes only the two characters; no terminating zero byte   */
/* is added.  No error is possible.                                */
/* --------------------------------------------------------------- */
void byte2hex(byte b, char *hex, flag upper) {
  static const char *uppers="0123456789ABCDEF";
  static const char *lowers="0123456789abcdef";
  const char *letters=(upper ? uppers : lowers);  // select table

  hex[0]=letters[(unsigned)b>>4];
  hex[1]=letters[b & 0x0f];
  } // byte2hex

/* --------------------------------------------------------------- */
/* hex2byte -- convert two hexadecimal characters to a byte        */
/*                                                                 */
/*   hex     -- points to two ASCII bytes                          */
/*   b       -- points to byte which will receive the converted    */
/*              hexadecimal; set to 0x00 if error                  */
/*   returns 0 if OK, -1 if error                                  */
/*                                                                 */
/* This accepts both uppercase and lowercase A->F and a->f.        */
/* --------------------------------------------------------------- */
int hex2byte(char *hex, byte *b) {
  char *end;        // end char +1
  *b=0;
  for (end=hex+2; hex<end; hex++) {
    *b<<=4;
    if (*hex>='0' && *hex<='9') *b|=(*hex-'0');
    else if (*hex>='A' && *hex<='F') *b|=(*hex-'A'+10);
    else if (*hex>='a' && *hex<='f') *b|=(*hex-'a'+10);
    else {*b=0x00; return -1;}
    }
  return 0;
  } // hex2byte

/* --------------------------------------------------------------- */
/* isqrtg -- return the square root of an int, with guess          */
/*                                                                 */
/*   num   -- the number to square root                            */
/*   guess -- is the number to converge from (e.g., last similar   */
/*            result); if < 1, isqrtg will make the first guess    */
/*                                                                 */
/*   returns the integer square root of num, such that             */
/*     result*result < num; if num is less than zero then 0 is     */
/*     returned                                                    */
/*                                                                 */
/* Note: the maximum input value is 2**31-1 (2147483647)           */
/* hence the maximum possible returned value is 46340.             */
/* --------------------------------------------------------------- */
// mfc, derived from PalmGlobe lsqrtg
int isqrtg(int num, int guess) {
  int new, old;

  if (num<9) {                /* special case low values for safety */
    if (num<=0) return 0;     /* special case + safety */
    if (num< 4) return 1;     /* 1, 2, 3 */
    return 2;}                /* 4-8 */

  /* Note: first guess has dramatic performance impact -- a value   */
  /* 1, or N/16, both doubled total GLOBE draw time.                */
  /* N/1024+1, and 32,000 both worked well.  Dynamic better.        */
  /* dynamically calculated first guess, which is (just) better.    */
  old=guess;                  /* The first guess is a given */
  if (old<=0) old=32001;
  for (;;) {                  /* forever */
    new=(signed int)((unsigned int)(num/old+old)>>1); /* new guess */
    if ((new-old)<2
     && (new-old)>-2) break;  /* Done! */
    old=new;
    }
  return new;
  } // isqrtg

/* --------------------------------------------------------------- */
/* i2str -- layout integer                                         */
/*                                                                 */
/*   string -- where to layout the integer                         */
/*   num    -- the integer to lay out (not 0x80000000)             */
/*   len    -- width of the field (string) to lay out into         */
/*                                                                 */
/*   returns 0 if OK, 1 if overflow (does not fit)                 */
/*                                                                 */
/* This writes num right-aligned and signed in the string field.   */
/* No terminating 0 is written.                                    */
/* --------------------------------------------------------------- */
int i2str(char *string, int num, int len) {
  char *c=string+len-1;                 // -> rightmost char
  char sign='+';
  int i;
  if (num<0) {num=-num; sign='-';}      // broken if 0x80000000
  for (i=0; i<len; i++) string[i]=' ';  // clear
  for (;;) {
    if (c==string) {                    // no room for sign
      for (i=0; i<len; i++) string[i]='*'; // show
      return 1;
      }
    *c=(num%10)+'0';
    num=num/10;
    c--;
    if (num==0) break;
    }
  *c=sign;
  return 0;
  } // i2str

/* --------------------------------------------------------------- */
/* memcmp -- compare bytes as per C library                        */
/* --------------------------------------------------------------- */
int memcmp(const void *s1, const void *s2, const size_t n) {
  uint u;
  const char *p1=s1;
  const char *p2=s2;
  for (u=0; u<n; u++, p1++, p2++) {
    if (*p1 < *p2) return -1;
    if (*p1 > *p2) return +1;
    }
  return 0;    // all bytes equal
  } // memcmp

/* --------------------------------------------------------------- */
/* memcpy -- copy bytes as per C library                           */
/* --------------------------------------------------------------- */
void *memcpy(void *s1, const void *s2, size_t n) {
  uint u;
  char *p1=s1;
  const char *p2=s2;
  for (u=0; u<n; u++, p1++, p2++) *p1=*p2;
  return s1;
  } // memcpy

/* --------------------------------------------------------------- */
/* memset -- set bytes as per C library                            */
/* --------------------------------------------------------------- */
void *memset(void *s, const int c, size_t n) {
  uint u;
  char *p=s;
  for (u=0; u<n; u++, p++) *p=c;
  return s;
  } // memset


